from __future__ import annotations

import dataclasses
import math

from ...common import encode_extended_float


@dataclasses.dataclass(frozen=True)
class Image:
    """A tensor's set of possible values

    See the methods below various convenience factories.
    """

    lower_bound: float = -math.inf
    """The image's smallest value (inclusive)"""

    upper_bound: float = math.inf
    """The image's largest value (inclusive)"""

    is_integral: bool = False
    """Whether the image only contain integers"""

    def render(self) -> str:
        if self.upper_bound == math.inf:
            if self.lower_bound == -math.inf:
                return "\\mathbb{Z}" if self.is_integral else "\\mathbb{R}"
            if self.lower_bound == 0:
                return "\\mathbb{N}" if self.is_integral else "\\mathbb{R}_+"
        lb = encode_extended_float(self.lower_bound)
        ub = encode_extended_float(self.upper_bound)
        if self.is_integral:
            if lb == 0 and ub == 1:
                return "\\{0, 1\\}"
            return f"\\{{{lb} \\ldots {ub}\\}}"
        return f"[{lb}, {ub}]"

    @classmethod
    def indicator(cls) -> Image:
        """Returns an image representing `{0, 1}` tensors"""
        return Image(lower_bound=0, upper_bound=1, is_integral=True)

    @classmethod
    def non_negative(cls, upper_bound=math.inf) -> Image:
        """Returns an image representing non-negative reals

        Args:
            upper_bound: Optional inclusive upper bound
        """
        return Image(lower_bound=0, upper_bound=upper_bound)

    @classmethod
    def non_positive(cls, lower_bound=-math.inf) -> Image:
        """Returns an image representing non-positive reals

        Args:
            lower_bound: Optional inclusive lower bound
        """
        return Image(lower_bound=lower_bound, upper_bound=0)

    @classmethod
    def unit(cls) -> Image:
        """Returns an image representing the `[0, 1]` segment"""
        return Image(lower_bound=0, upper_bound=1)

    @classmethod
    def natural(cls, upper_bound=math.inf) -> Image:
        """Returns an image representing natural (non-negative) integers

        Args:
            upper_bound: Optional inclusive upper bound
        """
        return Image(lower_bound=0, upper_bound=upper_bound, is_integral=True)

    @classmethod
    def integral(cls) -> Image:
        """Returns an image representing all integers"""
        return Image(is_integral=True)



from __future__ import annotations

import collections
import contextlib
import contextvars
import dataclasses
from typing import (
    Any,
    Generator,
    Iterable,
    KeysView,
    Literal,
    Mapping,
    Optional,
    Sequence,
)

from ...common import Label


Name = str


Environment = KeysView[Name]


@dataclasses.dataclass(eq=False, frozen=True)
class Identifier:
    def format(self) -> Name:
        """Generates the final name"""
        return _active_scope.get().format(self)


class GlobalIdentifier(Identifier):
    name: Optional[Name]


@dataclasses.dataclass(eq=False, frozen=True)
class DimensionIdentifier(GlobalIdentifier):
    name: Optional[Name]


TensorVariant = Literal["variable", "parameter"]


@dataclasses.dataclass(eq=False, frozen=True)
class TensorIdentifier(GlobalIdentifier):
    name: Optional[Name]
    variant: TensorVariant


@dataclasses.dataclass(eq=False, frozen=True)
class AliasIdentifier(GlobalIdentifier):
    name: Name


@dataclasses.dataclass(eq=False, frozen=True)
class QuantifierGroup:
    alias: AliasIdentifier
    subscripts: tuple[Any, ...]
    rank: int


class QuantifierIdentifier(Identifier):
    space: Any  # Space
    groups: Sequence[QuantifierGroup]
    name: Optional[Name]

    @classmethod
    def base(cls, space: Any) -> QuantifierIdentifier:
        return _BaseQuantifierIdentifier(space=space, groups=[])

    def grouped_within(self, group: QuantifierGroup) -> QuantifierIdentifier:
        return _BaseQuantifierIdentifier(
            space=self.space, groups=[group, *self.groups]
        ).named(self.name)

    def named(self, name: Optional[Name]) -> QuantifierIdentifier:
        if name is None or name == self.name:
            return self
        return _NamedQuantifierIdentifier(name=name, parent=self)

    @property
    def outer_group(self) -> Optional[QuantifierGroup]:
        return self.groups[0] if self.groups else None


@dataclasses.dataclass(eq=False, frozen=True)
class _BaseQuantifierIdentifier(QuantifierIdentifier):
    space: Any
    groups: Sequence[QuantifierGroup]
    name = None


@dataclasses.dataclass(eq=False, frozen=True)
class _NamedQuantifierIdentifier(QuantifierIdentifier):
    name: Name
    parent: QuantifierIdentifier

    @property
    def space(self) -> Any:
        return self.parent.space

    @property
    def groups(self) -> Sequence[QuantifierGroup]:  # type: ignore[override]
        return self.parent.groups


class IdentifierFormatter:
    def __init__(self, labels: Mapping[GlobalIdentifier, Label]) -> None:
        self._formatted: dict[GlobalIdentifier, bool] = {}
        self._labels = labels

    def formatted_globals(self) -> Sequence[GlobalIdentifier]:
        return list(self._formatted)

    def format(self, identifier: Identifier, env: Environment) -> Name:
        if isinstance(identifier, GlobalIdentifier):
            self._formatted[identifier] = True
            if identifier.name:
                return identifier.name
            label = self._labels[identifier]
            if isinstance(identifier, DimensionIdentifier):
                return self._format_dimension(label, env)
            elif isinstance(identifier, TensorIdentifier):
                if identifier.variant == "parameter":
                    return self._format_parameter(label, env)
                else:
                    return self._format_variable(label, env)
        elif isinstance(identifier, QuantifierIdentifier):
            return self.format_quantifier(identifier, env)
        raise TypeError(f"Unexpected identifier: {identifier}")

    def _format_dimension(self, label: Label, env: Environment) -> Name:
        raise NotImplementedError()

    def _format_parameter(self, label: Label, env: Environment) -> Name:
        raise NotImplementedError()

    def _format_variable(self, label: Label, env: Environment) -> Name:
        raise NotImplementedError()

    def format_quantifier(
        self, identifier: QuantifierIdentifier, env: Environment
    ) -> Name:
        raise NotImplementedError()


@dataclasses.dataclass
class _Scope:
    formatter: IdentifierFormatter
    environment: collections.ChainMap[Name, Identifier]  # To owning identifier
    quantifier_names: collections.ChainMap[QuantifierIdentifier, Name]
    global_names: dict[GlobalIdentifier, Name]

    def child(self) -> _Scope:
        return dataclasses.replace(
            self,
            environment=self.environment.new_child(),
            quantifier_names=self.quantifier_names.new_child(),
        )

    def format(self, identifier: Identifier) -> Name:
        if isinstance(identifier, GlobalIdentifier):
            name = self.global_names.get(identifier)
        else:
            if not isinstance(identifier, QuantifierIdentifier):
                raise TypeError(f"Unexpected identifier: {identifier}")
            name = self.quantifier_names.get(identifier)
            if name is None:
                raise Exception(f"Missing quantifier: {identifier}")
        if name is None:
            name = self.prepare(identifier)
        return name

    def prepare(self, identifier: Identifier) -> Name:
        name = self.formatter.format(identifier, self.environment.keys())
        owner = self.environment.get(name)
        if owner is not None and owner != identifier:
            raise Exception(f"Name collision: {name}")
        if isinstance(identifier, GlobalIdentifier):
            self.global_names[identifier] = name
            self.environment.maps[-1][name] = identifier
        else:
            if not isinstance(identifier, QuantifierIdentifier):
                raise TypeError(f"Unexpected identifier: {identifier}")
            self.environment[name] = identifier
            self.quantifier_names[identifier] = name
            while isinstance(identifier, _NamedQuantifierIdentifier):
                identifier = identifier.parent
                self.quantifier_names[identifier] = name
        return name


_active_scope: Any = contextvars.ContextVar("formatting_scope")


@contextlib.contextmanager
def global_formatting_scope(
    formatter: IdentifierFormatter,
    reserved: Optional[Mapping[Name, GlobalIdentifier]] = None,
) -> Generator[None, None, None]:
    scope = _active_scope.get(None)
    if scope:
        raise Exception("Identifier formatter already active")
    scope = _Scope(
        formatter=formatter,
        environment=collections.ChainMap(dict(reserved) if reserved else {}),
        quantifier_names=collections.ChainMap(),
        global_names={},
    )
    token = _active_scope.set(scope)
    try:
        yield
    finally:
        _active_scope.reset(token)


@contextlib.contextmanager
def local_formatting_scope(
    quantifiers: Iterable[QuantifierIdentifier],
) -> Generator[None, None, None]:
    scope = _active_scope.get(None)
    if not scope:
        raise Exception("Missing active formatter")
    child = scope.child()
    for quantifier in quantifiers:
        child.prepare(quantifier)
    token = _active_scope.set(child)
    try:
        yield
    finally:
        _active_scope.reset(token)



from __future__ import annotations

import collections
import dataclasses
import glob
import logging
import os
from typing import Any, ClassVar, Iterable, Optional, Sequence

from ..common import Json, Setting, default_api_url, json_dict
from ..executors import (
    Executor,
    JsonExecutorResult,
    default_executor,
)


_logger = logging.getLogger(__name__)


_DEFAULT_TITLE = "<inline>"


@dataclasses.dataclass(frozen=True)
class LocalSpecificationSource:
    text: str
    title: str = _DEFAULT_TITLE


@dataclasses.dataclass(frozen=True)
class LocalSpecificationAnnotation:
    issue_count: int
    issues: Sequence[Sequence[LocalSpecificationIssue]]
    counts: Sequence[collections.Counter]


@dataclasses.dataclass(frozen=True)
class LocalSpecificationIssue:
    start_offset: int
    end_offset: int
    message: str
    code: str


@dataclasses.dataclass(frozen=True)
class LocalSpecification:
    """
    A local specification

    This type of specification cannot be used to start attempts.
    """

    __executor: ClassVar[Executor] = default_executor(
        default_api_url(Setting.DOMAIN.read())
    )

    sources: Sequence[LocalSpecificationSource]

    annotation: Optional[LocalSpecificationAnnotation] = None

    @classmethod
    def set_default_executor(cls, executor: Executor) -> None:
        cls.__executor = executor

    @classmethod
    def inline(cls, *texts: str) -> LocalSpecification:
        """Creates a local specification from strings"""
        sources = [LocalSpecificationSource(s) for s in texts]
        return LocalSpecification(sources)

    @classmethod
    def globs(
        cls, *likes: str, root: Optional[str] = None
    ) -> LocalSpecification:
        """Creates a local specification from file globs

        As a convenience the root can be a file's name, in which case it will
        be interpreted as its parent directory (this is typically handy when
        used with `__file__`).
        """
        if root:
            root = os.path.realpath(root)
            if os.path.isfile(root):
                root = os.path.dirname(root)
        sources: list[LocalSpecificationSource] = []
        for like in likes:
            for path in glob.iglob(like, root_dir=root, recursive=True):
                if root:
                    path = os.path.join(root, path)
                with open(path, "r", encoding="utf-8") as reader:
                    sources.append(
                        LocalSpecificationSource(
                            text=reader.read(),
                            title=path,
                        )
                    )
        return LocalSpecification(sources)

    async def annotated(
        self,
        ignore_codes: Optional[Iterable[str]] = None,
        executor: Optional[Executor] = None,
    ) -> LocalSpecification:
        sources = self.sources
        codes = set(ignore_codes or [])

        if not executor:
            executor = self.__executor
        async with executor.execute(
            result_type=JsonExecutorResult,
            url="/sources/parse",
            method="POST",
            json_data=json_dict(sources=[s.text for s in sources]),
        ) as res:
            data = res.json_data()
            counts: Sequence[Any] = [collections.Counter() for _ in sources]
            for s in data["slices"]:
                category = s["definition"]["category"].lower()
                counts[s["index"]][category] += 1
            issues: Sequence[Any] = [[] for _ in sources]
            for e in data["errors"]:
                if e["code"] in codes:
                    continue
                issues[e["index"]].append(_issue_from_json(e))
            annotation = LocalSpecificationAnnotation(
                issue_count=len(data["errors"]),
                issues=issues,
                counts=counts,
            )
        return dataclasses.replace(self, annotation=annotation)

    def _repr_markdown_(self) -> str:
        annotation = self.annotation
        sources = self.sources
        if annotation:
            for i, issues in enumerate(annotation.issues):
                if not issues:
                    continue
                messages = [f"  * [{i.code}] {i.message}" for i in issues]
                _logger.error(
                    "%s issue(s) in %s specification:\n%s",
                    len(issues),
                    sources[i].title,
                    "\n".join(messages),
                )
        return "\n\n---\n\n".join(
            _source_details(
                s,
                annotation.counts[i] if annotation else None,
                annotation.issues[i] if annotation else [],
                start_open=i == 0,
            )
            for i, s in enumerate(self.sources)
        )


def _source_details(
    source: LocalSpecificationSource,
    counts: Optional[collections.Counter],
    issues: Sequence[LocalSpecificationIssue],
    start_open=False,
) -> str:
    lines = ["<details open>" if start_open else "<details>"]
    if source.title:
        lines.append('<summary style="cursor:pointer;">&#9656; ')
        summary = source.title
        if counts:
            parens = ", ".join(f"{k}: {v}" for k, v in counts.most_common())
            summary += f" ({parens})"
        lines.append(summary)
        lines.append("</summary>\n")
    lines.append(_colorize(source.text, issues))
    lines.append("</details>")
    return "\n".join(lines)


def _issue_from_json(data: Json) -> LocalSpecificationIssue:
    rg = data["range"]
    return LocalSpecificationIssue(
        message=data["message"],
        start_offset=rg["start"]["offset"],
        end_offset=rg["end"]["offset"],
        code=data["code"],
    )


def _colorize(text: str, issues: Sequence[LocalSpecificationIssue]) -> str:
    if not issues:
        return text

    by_start = sorted(issues, key=lambda i: i.start_offset)
    cutoffs: list[int] = []
    end = 0
    for issue in by_start:
        if issue.code == "ERR_INVALID_SYNTAX":
            # These issues can't be easily colorized as they span name
            # boundaries
            continue
        if issue.start_offset <= end:
            end = max(issue.end_offset + 1, end)
        else:
            cutoffs.append(end)
            cutoffs.append(issue.start_offset)
            end = issue.end_offset + 1
    cutoffs.append(end)
    cutoffs.append(len(text))

    ret: list[str] = []
    for i, piece in enumerate(text[m:n] for m, n in zip(cutoffs, cutoffs[1:])):
        is_error = i % 2 == 1
        if is_error:
            ret.append(f"\\color{{orange}}{{{piece}}}")
        else:
            ret.append(piece)
    return "".join(ret)



from __future__ import annotations

import backoff
from datetime import datetime, timezone
import json
import humanize
import logging
from typing import cast, Mapping, Optional, Union

from ..common import (
    Json,
    Setting,
    default_api_url,
    default_hub_url,
    format_percent,
    json_dict,
)
from ..data.attempts import (
    Attempt,
    attempt_from_graphql,
    AttemptNotification,
    notification_from_graphql,
)
from ..data.outcomes import (
    CancelledOutcome,
    failed_outcome_from_graphql,
    FeasibleOutcome,
    feasible_outcome_from_graphql,
    InfeasibleOutcome,
    Outcome,
    outcome_status,
    UnboundedOutcome,
    UnexpectedOutcomeError,
)
from ..data.outlines import Label, Outline, outline_from_json
from ..data.solves import (
    SolveInputs,
    SolveOptions,
    SolveOutputs,
    SolveResponse,
    SolveStrategy,
    solve_options_to_json,
    solve_response_from_json,
    solve_strategy_to_json,
    solve_summary_from_json,
)
from ..data.tensors import DimensionArgument, TensorArgument
from ..executors import (
    default_executor,
    Executor,
    JsonExecutorResult,
    JsonSeqExecutorResult,
    PlainTextExecutorResult,
)
from ..specifications import (
    FormulationSpecification,
    RemoteSpecification,
    Specification,
)
from ..transformations import Transformation
from .common import (
    feasible_outcome_details,
    log_progress,
    OutlineGenerator,
    SolveInputsBuilder,
)


_logger = logging.getLogger(__name__)


class Client:
    """Opvious API client"""

    def __init__(self, executor: Executor, api_url: str, hub_url: str):
        self._executor = executor
        self._api_url = api_url
        self._hub_url = hub_url

    def __repr__(self) -> str:
        fields = [
            f"executor_class={self._executor.__class__.__name__}",
            f"api_url={json.dumps(self._api_url)}",
        ]
        return f"<opvious.Client {' '.join(fields)}>"

    @classmethod
    def from_token(cls, token: str, domain: Optional[str] = None) -> Client:
        """
        Creates a client from an API token

        Args:
            token: API token. You can use an empty token to create an
                unauthenticated client with limited functionality.
            domain: API domain. You should only need to set this if you are
                using a self-hosted cluster. Defaults to the default production
                endpoint.
        """
        token = token.strip()
        authorization = None
        if token:
            authorization = token if " " in token else f"Bearer {token}"
        api_url = default_api_url(domain)
        return Client(
            executor=default_executor(
                root_url=api_url,
                authorization=authorization,
            ),
            api_url=api_url,
            hub_url=default_hub_url(domain),
        )

    @classmethod
    def from_environment(
        cls, env: Optional[dict[str, str]] = None, require_authenticated=False
    ) -> Client:
        """Creates a client from environment variables

        Args:
            env: Environment, defaults to `os.environ`.
            require_authenticated: Throw if the environment does not include a
                valid API token.
        """
        token = Setting.TOKEN.read(env)
        if not token and require_authenticated:
            raise Exception(
                f"Missing or empty {Setting.TOKEN.value} environment variable"
            )
        return Client.from_token(token=token, domain=Setting.DOMAIN.read(env))

    @property
    def authenticated(self) -> bool:
        """Returns true if the client was created with a non-empty API token"""
        return self._executor.authenticated

    @property
    def executor(self) -> Executor:
        """Returns the client's underlying executor"""
        return self._executor

    async def _prepare_solve(
        self,
        specification: Specification,
        parameters: Optional[Mapping[Label, TensorArgument]] = None,
        dimensions: Optional[Mapping[Label, DimensionArgument]] = None,
        transformations: Optional[list[Transformation]] = None,
        strategy: Optional[SolveStrategy] = None,
        options: Optional[SolveOptions] = None,
    ) -> tuple[Json, Outline]:
        """Generates solve candidate and final outline."""
        # First we fetch the outline to validate/coerce inputs later on
        if isinstance(specification, FormulationSpecification):
            outline_generator, tag = await OutlineGenerator.formulation(
                executor=self._executor,
                specification=specification,
            )
            formulation = json_dict(
                name=specification.formulation_name,
                specification_tag_name=tag,
            )
        else:
            if isinstance(specification, RemoteSpecification):
                sources = await specification.fetch_sources(self._executor)
            else:
                sources = [s.text for s in specification.sources]
            formulation = json_dict(sources=sources)
            outline_generator = await OutlineGenerator.sources(
                executor=self._executor, sources=sources
            )

        # Then we apply any transformations and refresh the outline if needed
        for tf in transformations or []:
            outline_generator.add_transformation(tf)
        outline, transformation_data = await outline_generator.generate()

        # Then we assemble the inputs
        builder = SolveInputsBuilder(outline=outline)
        if dimensions:
            for label, dim in dimensions.items():
                builder.set_dimension(label, dim)
        if parameters:
            for label, param in parameters.items():
                builder.set_parameter(label, param)
        inputs = builder.build()
        _logger.info(
            "Validated inputs. [parameters=%s]",
            builder.parameter_entry_count,
        )

        # Finally we put everything together
        candidate = json_dict(
            formulation=formulation,
            inputs=json_dict(
                dimensions=inputs.raw_dimensions,
                parameters=inputs.raw_parameters,
            ),
            transformations=transformation_data,
            strategy=solve_strategy_to_json(strategy, outline),
            options=solve_options_to_json(options),
        )
        return (candidate, outline)

    async def inspect_solve_instructions(
        self,
        specification: Specification,
        parameters: Optional[Mapping[Label, TensorArgument]] = None,
        dimensions: Optional[Mapping[Label, DimensionArgument]] = None,
        transformations: Optional[list[Transformation]] = None,
        strategy: Optional[SolveStrategy] = None,
        options: Optional[SolveOptions] = None,
    ) -> str:
        """Returns the model's representation in `LP format`_

        The arguments below are identical to :meth:`.Client.run_solve`, making
        it easy to swap one call for another when debugging.

        Args:
            specification: :ref:`Model specification <Specifications>`
            parameters: Input data, keyed by parameter label. Values may be any
                value accepted by :meth:`.Tensor.from_argument` and must match
                the corresponding parameter's definition.
            dimensions: Dimension items, keyed by dimension label. If omitted,
                these will be automatically inferred from the parameters.
            transformations: :ref:`Model transformations`
            strategy: :ref:`Multi-objective strategy <Multi-objective
                strategies>`
            options: Solve options

        The LP formatted output will be fully annotated with matching keys and
        labels:

        .. code::

            minimize
              +1 inventory$1 \\ [day=0]
              +1 inventory$2 \\ [day=1]
              +1 inventory$3 \\ [day=2]
              \\ ...
            subject to
             inventoryPropagation$1: \\ [day=1]
              +1 inventory$1 \\ [day=1]
              -1 inventory$2 \\ [day=0]
              -1 production$1 \\ [day=1]
              = -29
             inventoryPropagation$2: \\ [day=2]
              -1 inventory$1 \\ [day=1]
              +1 inventory$3 \\ [day=2]
              -1 production$2 \\ [day=2]
              = -36
             \\ ...

        .. _LP format: https://web.mit.edu/lpsolve/doc/CPLEX-format.htm
        """
        candidate, _outline = await self._prepare_solve(
            specification=specification,
            parameters=parameters,
            dimensions=dimensions,
            transformations=transformations,
            options=options,
        )
        async with self._executor.execute(
            result_type=PlainTextExecutorResult,
            url="/solves/inspect/instructions",
            method="POST",
            json_data=json_dict(candidate=candidate),
        ) as res:
            lines = []
            async for line in res.lines():
                if line.startswith("\\"):
                    _logger.debug(line[2:].strip())
                else:
                    lines.append(line)
            return "".join(lines)

    async def run_solve(
        self,
        specification: Specification,
        parameters: Optional[Mapping[Label, TensorArgument]] = None,
        dimensions: Optional[Mapping[Label, DimensionArgument]] = None,
        transformations: Optional[list[Transformation]] = None,
        strategy: Optional[SolveStrategy] = None,
        options: Optional[SolveOptions] = None,
        assert_feasible=False,
        prefer_streaming=True,
    ) -> SolveResponse:
        """Solves an optimization problem remotely

        Inputs will be validated before being sent to the API for solving.

        Args:
            specification: :ref:`Model specification <Specifications>`
            parameters: Input data, keyed by parameter label. Values may be any
                value accepted by :meth:`.Tensor.from_argument` and must match
                the corresponding parameter's definition.
            dimensions: Dimension items, keyed by dimension label. If omitted,
                these will be automatically inferred from the parameters.
            transformations: :ref:`Model transformations`
            strategy: :ref:`Multi-objective strategy <Multi-objective
                strategies>`
            options: Solve options
            assert_feasible: Throw if the final outcome was not feasible
            prefer_streaming: Show real time progress notifications when
                possible

        The returned response exposes both metadata (status, objective value,
        etc.) and solution data (if the solve was feasible):

        .. code:: python

            response = await client.run_solve(
                specification=opvious.RemoteSpecification.example(
                    "porfolio-selection"
                ),
                parameters={
                    "covariance": {
                        ("AAPL", "AAPL"): 0.2,
                        ("AAPL", "MSFT"): 0.1,
                        ("MSFT", "AAPL"): 0.1,
                        ("MSFT", "MSFT"): 0.25,
                    },
                    "expectedReturn": {
                        "AAPL": 0.15,
                        "MSFT": 0.2,
                    },
                    "desiredReturn": 0.1,
                },
                assert_feasible=True,  # Throw if not feasible
            )

            # Metadata is available on `outcome`
            print(f"Objective value: {response.outcome.objective_value}")

            # Solution data is available via `outputs`
            optimal_allocation = response.outputs.variable("allocation")


        See also :meth:`.Client.start_attempt` for an alternative for
        long-running solves.
        """
        candidate, outline = await self._prepare_solve(
            specification=specification,
            parameters=parameters,
            dimensions=dimensions,
            transformations=transformations,
            strategy=strategy,
            options=options,
        )

        if prefer_streaming and self._executor.supports_streaming:
            summary = None
            response_json = None
            async with self._executor.execute(
                result_type=JsonSeqExecutorResult,
                url="/solves/run",
                method="POST",
                json_data=json_dict(candidate=candidate),
            ) as res:
                async for data in res.json_seq_data():
                    kind = data["kind"]
                    if kind == "reifying":
                        progress = data["progress"]
                        if progress["kind"] == "constraint":
                            summary = progress["summary"]
                            _logger.debug(
                                "Reified constraint %r. [columns=%s, rows=%s]",
                                summary["label"],
                                summary["columnCount"],
                                summary["rowCount"],
                            )
                    elif kind == "reified":
                        summary = solve_summary_from_json(data["summary"])
                        _logger.info(
                            "Solving problem... [columns=%s, rows=%s]",
                            summary.column_count,
                            summary.row_count,
                        )
                    elif kind == "solving":
                        log_progress(_logger, data["progress"])
                    elif kind == "solved":
                        _logger.debug("Downloaded outputs.")
                        response_json = data
                    elif kind == "error":
                        message = data["error"]["message"]
                        raise Exception(f"Solve failed: {message}")
                    else:
                        raise Exception(
                            f"Unexpected response: {json.dumps(data)}"
                        )
            if not summary or not response_json:
                raise Exception("Streaming solve terminated early")
            response = solve_response_from_json(
                outline=outline,
                response_json=response_json,
                summary=summary,
            )
        else:
            async with self._executor.execute(
                result_type=JsonExecutorResult,
                url="/solves/run",
                method="POST",
                json_data=json_dict(candidate=candidate),
            ) as res:
                response = solve_response_from_json(
                    outline=outline,
                    response_json=res.json_data(),
                )

        outcome = response.outcome
        if isinstance(outcome, FeasibleOutcome):
            details = feasible_outcome_details(outcome)
            _logger.info(
                "Solve completed with status %s.%s",
                response.status,
                f" [{details}]" if details else "",
            )
        elif assert_feasible:
            raise UnexpectedOutcomeError(response.outcome)
        else:
            _logger.info("Solve completed with status %s.", response.status)

        return response

    def _attempt_url(self, uuid) -> str:
        return self._hub_url + "/attempts/" + uuid

    async def start_attempt(
        self,
        specification: Union[str, FormulationSpecification],
        parameters: Optional[Mapping[Label, TensorArgument]] = None,
        dimensions: Optional[Mapping[Label, DimensionArgument]] = None,
        transformations: Optional[list[Transformation]] = None,
        strategy: Optional[SolveStrategy] = None,
        options: Optional[SolveOptions] = None,
    ) -> Attempt:
        """Starts a new asynchronous remote solve attempt

        Inputs will be validated locally before the request is sent to the API.
        From then on, he attempt will be queued and begin solving start as soon
        as enough capacity is available.

        Args:
            specification: Model :class:`.FormulationSpecification` or
                formulation name
            parameters: Input data, keyed by parameter label. Values may be any
                value accepted by :meth:`.Tensor.from_argument` and must match
                the corresponding parameter's definition.
            dimensions: Dimension items, keyed by dimension label. If omitted,
                these will be automatically inferred from the parameters.
            transformations: :ref:`Model transformations`
            strategy: :ref:`Multi-objective strategy <Multi-objective
                strategies>`
            options: Solve options

        The returned :class:`Attempt` instance can be used to:

        + track progress via :meth:`Client.poll_attempt`,
        + retrieve inputs via :meth:`Client.fetch_attempt_inputs`,
        + retrieve outputs via :meth:`Client.fetch_attempt_outputs` (after
          successful completion).

        As a convenience, :meth:`Client.wait_for_outcome` allows polling an
        attempt until until it completes, backing off exponentially between
        each poll:

        .. code-block:: python

            # Queue a new Sudoku solve attempt
            attempt = await client.start_attempt(
                specification="sudoku",
                parameters={"hints": [(0, 0, 3), (1, 1, 5)]},
            )

            # Wait for the attempt to complete
            await client.wait_for_outcome(
                attempt,
                assert_feasible=True  # Throw if not feasible
            )

            # Fetch the solution's data
            output_data = await client.fetch_attempt_outputs(attempt)

            # Get a parsed variable as a dataframe
            decisions = output_data.variable("decisions")

        See also :meth:`.Client.run_solve` for an alternative for short solves.
        """
        if isinstance(specification, str):
            specification = FormulationSpecification(
                formulation_name=specification
            )
        candidate, outline = await self._prepare_solve(
            specification=specification,
            parameters=parameters,
            dimensions=dimensions,
            transformations=transformations,
            strategy=strategy,
            options=options,
        )
        async with self._executor.execute(
            result_type=JsonExecutorResult,
            url="/attempts/start",
            method="POST",
            json_data=json_dict(candidate=candidate),
        ) as res:
            uuid = res.json_data()["uuid"]
        return Attempt(
            uuid=uuid,
            started_at=datetime.now(timezone.utc),
            outline=outline,
            url=self._attempt_url(uuid),
        )

    async def load_attempt(self, uuid: str) -> Optional[Attempt]:
        """Loads an existing attempt

        Args:
            uuid: The target attempt's ID
        """
        data = await self._executor.execute_graphql_query(
            query="@FetchAttempt",
            variables=json_dict(uuid=uuid),
        )
        attempt = data["attempt"]
        if not attempt:
            return None
        return attempt_from_graphql(
            data=attempt,
            outline=outline_from_json(attempt["outline"]),
            url=self._attempt_url(uuid),
        )

    async def cancel_attempt(self, uuid: str) -> bool:
        """Cancels a running attempt

        This method will throw if the attempt does not exist or is not pending
        anymore.

        Args:
            uuid: The target attempt's ID
        """
        data = await self._executor.execute_graphql_query(
            query="@CancelAttempt",
            variables=json_dict(uuid=uuid),
        )
        return bool(data["cancelAttempt"])

    async def poll_attempt(
        self, attempt: Attempt
    ) -> Union[AttemptNotification, Outcome]:
        """Polls an attempt for its outcome or latest progress notification

        Args:
            attempt: The target attempt
        """
        data = await self._executor.execute_graphql_query(
            query="@PollAttempt",
            variables=json_dict(uuid=attempt.uuid),
        )
        attempt_data = data["attempt"]
        status = attempt_data["status"]
        if status == "PENDING":
            edges = attempt_data["notifications"]["edges"]
            return notification_from_graphql(
                dequeued=bool(attempt_data["dequeuedAt"]),
                data=edges[0]["node"] if edges else None,
            )
        if status == "CANCELLED":
            return cast(Outcome, CancelledOutcome())
        if status == "INFEASIBLE":
            return InfeasibleOutcome()
        if status == "UNBOUNDED":
            return UnboundedOutcome()
        outcome = attempt_data["outcome"]
        if status == "ERRORED":
            return failed_outcome_from_graphql(outcome)
        if status == "FEASIBLE" or status == "OPTIMAL":
            return feasible_outcome_from_graphql(outcome)
        raise Exception(f"Unexpected status {status}")

    @backoff.on_predicate(
        backoff.fibo,
        lambda ret: ret is None,
        max_value=90,
        logger=None,
    )
    async def _track_attempt(self, attempt: Attempt) -> Optional[Outcome]:
        ret = await self.poll_attempt(attempt)
        if isinstance(ret, AttemptNotification):
            delta = datetime.now(timezone.utc) - attempt.started_at
            elapsed = humanize.naturaldelta(delta, minimum_unit="milliseconds")
            if ret.dequeued:
                details = [f"elapsed={elapsed}"]
                if ret.relative_gap is not None:
                    details.append(f"gap={format_percent(ret.relative_gap)}")
                if ret.cut_count is not None:
                    details.append(f"cuts={ret.cut_count}")
                if ret.lp_iteration_count is not None:
                    details.append(f"iterations={ret.lp_iteration_count}")
                _logger.info("Attempt is running... [%s]", ", ".join(details))
            else:
                _logger.info("Attempt is queued... [elapsed=%s]", elapsed)
            return None
        return ret

    async def wait_for_outcome(
        self,
        attempt: Attempt,
        assert_feasible=False,
    ) -> Outcome:
        """Waits for the attempt to complete and returns its outcome

        This method emits real-time progress messages as INFO logs.

        Args:
            attempt: The target attempt
            assert_feasible: Throw if the final outcome was not feasible
        """
        outcome = await self._track_attempt(attempt)
        if not outcome:
            raise Exception("Missing outcome")
        status = outcome_status(outcome)
        if isinstance(outcome, FeasibleOutcome):
            details = feasible_outcome_details(outcome)
            _logger.info(
                "Attempt completed with status %s.%s",
                status,
                f" [{details}]" if details else "",
            )
        elif assert_feasible:
            raise UnexpectedOutcomeError(outcome)
        else:
            _logger.info("Attempt completed with status %s.", status)
        return outcome

    async def fetch_attempt_inputs(self, attempt: Attempt) -> SolveInputs:
        """Retrieves an attempt's inputs

        Args:
            attempt: The target attempt
        """
        async with self._executor.execute(
            result_type=JsonExecutorResult,
            url=f"/attempts/{attempt.uuid}/inputs",
        ) as res:
            data = res.json_data()
        return SolveInputs(
            outline=attempt.outline,
            raw_parameters=data["parameters"],
            raw_dimensions=data["dimensions"],
        )

    async def fetch_attempt_outputs(self, attempt: Attempt) -> SolveOutputs:
        """Retrieves a successful attempt's outputs

        This method will throw if the attempt did not have a feasible solution.

        Args:
            attempt: The target attempt
        """
        async with self._executor.execute(
            result_type=JsonExecutorResult,
            url=f"/attempts/{attempt.uuid}/outputs",
        ) as res:
            data = res.json_data()
        return SolveOutputs(
            outline=attempt.outline,
            raw_variables=data["variables"],
            raw_constraints=data["constraints"],
        )



from __future__ import annotations

import dataclasses
from datetime import datetime
from typing import Any, Optional

from .outlines import Outline
from .tensors import Value


@dataclasses.dataclass(frozen=True)
class Attempt:
    """Queueable optimization attempt

    New attempts are started via :meth:`.Client.start_attempt`, existing
    attempts can be retrieved from their UUID via :meth:`.Client.load_attempt`.
    """

    uuid: str
    """The attempt's unique identifier"""

    started_at: datetime
    """The time the attempt was created"""

    outline: Outline = dataclasses.field(repr=False)
    """The specification outline corresponding to this attempt"""

    url: str = dataclasses.field(repr=False)
    """The optimimization hub URL for this attempt"""


def attempt_from_graphql(data: Any, outline: Outline, url: str) -> Attempt:
    return Attempt(
        uuid=data["uuid"],
        started_at=datetime.fromisoformat(data["startedAt"]),
        outline=outline,
        url=url,
    )


@dataclasses.dataclass(frozen=True)
class AttemptNotification:
    """Attempt progress update notification"""

    dequeued: bool
    """Whether the attempt has already been dequeued"""

    relative_gap: Optional[Value]
    """The latest relative gap"""

    lp_iteration_count: Optional[int]
    """The latest LP iteration count"""

    cut_count: Optional[int]
    """The latest cut count"""


def notification_from_graphql(
    dequeued: bool, data: Any = None
) -> AttemptNotification:
    return AttemptNotification(
        dequeued=dequeued,
        relative_gap=data["relativeGap"] if data else None,
        lp_iteration_count=data["lpIterationCount"] if data else None,
        cut_count=data["cutCount"] if data else None,
    )



from typing import Union

from .local import (
    LocalSpecification,
    LocalSpecificationIssue,
    LocalSpecificationSource,
)
from .external import FormulationSpecification, RemoteSpecification


Specification = Union[
    LocalSpecification,
    RemoteSpecification,
    FormulationSpecification,
]


__all__ = [
    "FormulationSpecification",
    "LocalSpecification",
    "LocalSpecificationIssue",
    "LocalSpecificationSource",
    "RemoteSpecification",
    "Specification",
]



import enum
from importlib import metadata
import math
import os
from typing import Any, Optional, Union, cast
import urllib.parse


try:
    __version__ = metadata.version(__package__)
except metadata.PackageNotFoundError:
    __version__ = ""


del metadata


# Configuration


_DEFAULT_DOMAIN = "beta.opvious.io"


class Setting(enum.Enum):
    """Client configuration environment variables"""

    TOKEN = ("OPVIOUS_TOKEN", "")
    DOMAIN = ("OPVIOUS_DOMAIN", _DEFAULT_DOMAIN)

    def read(self, env: Optional[dict[str, str]] = None) -> str:
        """Read the setting's current value or default if missing

        Args:
            env: Environment, defaults to `os.environ`.
        """
        if env is None:
            env = cast(Any, os.environ)
        name, default_value = self.value
        return env.get(name) or default_value


def default_api_url(domain: Optional[str] = None) -> str:
    return f"https://api.{domain or _DEFAULT_DOMAIN}"


def default_hub_url(domain: Optional[str] = None) -> str:
    return f"https://hub.{domain or _DEFAULT_DOMAIN}"


# Formatting


Label = str


def format_percent(val):
    if val == "Infinity":
        return "inf"
    return f"{int(val * 10_000) / 100}%"


def is_url(s: str) -> bool:
    """Checks if a string is a URL."""
    try:
        res = urllib.parse.urlparse(s)
        return bool(res.scheme and res.netloc)
    except ValueError:
        return False


def to_camel_case(s: str) -> str:
    if "_" not in s:
        return s
    return "".join(
        p.capitalize() if i else p for i, p in enumerate(s.split("_"))
    )


def untuple(t: tuple[Any, ...]) -> Any:
    return t[0] if len(t) == 1 else t


# JSON utilities


Json = Any


ExtendedFloat = Union[float, str]


def encode_extended_float(val: ExtendedFloat) -> Json:
    if val == math.inf:
        return "Infinity"
    elif val == -math.inf:
        return "-Infinity"
    return val


def decode_extended_float(val: ExtendedFloat) -> Json:
    if val == "Infinity":
        return math.inf
    elif val == "-Infinity":
        return -math.inf
    return val


def json_dict(**kwargs) -> Json:
    """Strips keys with None values and encodes infinity values"""
    data = {}
    for key, val in kwargs.items():
        if val is None:
            continue
        json_key = to_camel_case(key)
        if isinstance(val, float):
            data[json_key] = encode_extended_float(val)
        else:
            data[json_key] = val
    return data



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['opvious',
 'opvious.client',
 'opvious.data',
 'opvious.executors',
 'opvious.specifications',
 'opvious.specifications.model']

package_data = \
{'': ['*']}

install_requires = \
['backoff>=2.2,<3.0', 'humanize>=4.4.0,<5.0.0', 'pandas>=1.4,<2.0']

extras_require = \
{'aio': ['aiohttp>=3.8,<4.0', 'Brotli>=1.0.9,<2.0.0']}

setup_kwargs = {
    'name': 'opvious',
    'version': '0.12.0rc11',
    'description': 'Opvious Python SDK',
    'long_description': '# Opvious Python SDK  [![CI](https://github.com/opvious/sdk.py/actions/workflows/ci.yml/badge.svg)](https://github.com/opvious/sdk.py/actions/workflows/ci.yml) [![Pypi badge](https://badge.fury.io/py/opvious.svg)](https://pypi.python.org/pypi/opvious/)\n\nAn optimization SDK for solving linear, mixed-integer, and quadratic models\n\n```python\nimport opvious\n\nclient = opvious.Client.from_environment()\n\n# Solve a portfolio selection optimization model\nresponse = await client.run_solve(\n    specification=opvious.LocalSpecification.inline(\n        r"""\n        We find an allocation of assets which minimizes risk while satisfying\n        a minimum expected return:\n\n        + A collection of assets: $\\S^d_{asset}: A$\n        + Covariances: $\\S^p_{covariance}: c \\in \\mathbb{R}^{A \\times A}$\n        + Expected return: $\\S^p_{expectedReturn}: m \\in \\mathbb{R}^A$\n        + Minimum desired return: $\\S^p_{desiredReturn}: r \\in \\mathbb{R}$\n\n        The only output is the allocation per asset\n        $\\S^v_{allocation}: \\alpha \\in [0,1]^A$ chosen to minimize risk:\n        $\\S^o_{risk}: \\min \\sum_{a, b \\in A} c_{a,b} \\alpha_a \\alpha_b$.\n\n        Subject to the following constraints:\n\n        + $\\S^c_{atLeastMinimumReturn}: \\sum_{a \\in A} m_a \\alpha_a \\geq r$\n        + $\\S^c_{totalAllocation}: \\sum_{a \\in A} \\alpha_a = 1$\n        """\n    ),\n    parameters={\n        "covariance": {\n            ("AAPL", "AAPL"): 0.08,\n            # ...\n        },\n        "expectedReturn": {\n            "AAPL": 0.07,\n            # ..\n        },\n        "desiredReturn": 0.05,\n    },\n    assert_feasible=True,\n)\n\noptimal_allocation = response.outputs.variable("allocation")\n```\n\nTake a look at https://opvious.readthedocs.io for the full documentation or\n[these notebooks][notebooks] to see it in action.\n\n[notebooks]: https://github.com/opvious/examples/tree/main/notebooks\n',
    'author': 'Opvious Engineering',
    'author_email': 'oss@opvious.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/opvious/sdk.py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.9,<3.12',
}


setup(**setup_kwargs)



from __future__ import annotations

import collections
import dataclasses
import logging
from typing import Any, Iterable, Mapping, Optional, Sequence

from ...common import Label, to_camel_case
from ..local import LocalSpecification, LocalSpecificationSource
from .ast import Space
from .identifiers import (
    Environment,
    GlobalIdentifier,
    IdentifierFormatter,
    Name,
    QuantifierIdentifier,
    global_formatting_scope,
)


_logger = logging.getLogger(__name__)


class Definition:
    """Internal model definition"""

    @property
    def label(self) -> Optional[Label]:
        raise NotImplementedError()

    @property
    def identifier(self) -> Optional[GlobalIdentifier]:
        raise NotImplementedError()

    def render_statement(self, label: Label, model: Any) -> Optional[str]:
        raise NotImplementedError()


class ModelFragment:
    """Model partial"""


@dataclasses.dataclass(frozen=True)
class _Statement:
    label: Label
    definition: Definition
    fragment: Optional[ModelFragment]
    model: Model

    def render(self) -> Optional[str]:
        owner = self.fragment or self.model
        return self.definition.render_statement(self.label, owner)


@dataclasses.dataclass(frozen=True)
class _Relabeled(ModelFragment):
    fragment: ModelFragment
    labels: Mapping[str, Label]


def relabel(fragment: ModelFragment, **kwargs: Label) -> ModelFragment:
    """Updates a fragment's definitions' labels"""
    return _Relabeled(fragment, kwargs)


class _ModelVisitor:
    def __init__(self) -> None:
        self.statements: list[_Statement] = []
        self._visited: set[int] = set()  # Fragment IDs

    def visit(
        self,
        model: Model,
        prefix: Optional[str] = None,
    ) -> None:
        model_id = id(model)
        if model_id in self._visited:
            return
        self._visited.add(model_id)

        self._visit_fragment(model, None, [prefix] if prefix else [])
        for dep in model.dependencies:
            self.visit(dep)

    def _visit_fragment(
        self,
        model: Model,
        frag: Optional[ModelFragment],
        prefix: Sequence[str],
    ) -> None:
        obj = frag or model

        labels: dict[str, Label] = {}
        while isinstance(obj, _Relabeled):
            labels.update(obj.labels)
            obj = obj.fragment

        attrs: dict[str, Any] = {}
        for cls in reversed(obj.__class__.__mro__[1:]):
            attrs.update(cls.__dict__)
        attrs.update(obj.__dict__)
        attrs.update(obj.__class__.__dict__)

        path = [*prefix, ""]
        for attr, value in attrs.items():
            if attr.startswith("_"):
                continue
            path[-1] = attr
            if isinstance(value, property):
                value = value.fget
            if not isinstance(value, Definition):
                if isinstance(value, ModelFragment):
                    self._visit_fragment(model, value, path)
                continue
            label = (
                labels.get(attr)
                or value.label
                or to_camel_case("_".join(path))
            )
            self.statements.append(_Statement(label, value, frag, model))


class Model:
    """An optimization model

    Args:
        dependencies: Optional list of models upon which this model's
            definitions depend. Dependencies' definitions will be automatically
            added when generating this model's specification.

    Toy example for the set cover problem:

    .. code-block:: python

        class SetCover(Model):
            sets = Dimension()
            vertices = Dimension()
            covers = Parameter(sets, vertices, image=indicator())
            used = Variable(sets, image=indicator())

            @constraint
            def all_covered(self):
                for v in self.vertices:
                    count = total(
                        self.used(s) * self.covers(s, v)
                        for s in self.sets
                    )
                    yield count >= 1

            @objective
            def minimize_used(self):
                return total(self.used(s) for s in self.sets)
    """

    __dependencies: Optional[Sequence[Model]] = None
    __prefix: Optional[str] = None
    __title: Optional[str] = None

    def __init__(
        self,
        dependencies: Optional[Iterable[Model]] = None,
        prefix: Optional[Label] = None,
        title: Optional[str] = None,
    ):
        self.__dependencies = list(dependencies) if dependencies else None
        self.__prefix = prefix
        self.__title = title

    @property
    def title(self) -> str:
        return self.__title or f"<code>{self.__class__.__name__}</code>"

    @property
    def dependencies(self) -> Sequence[Model]:
        """The model's dependencies"""
        return self.__dependencies or []

    async def compile_specification(
        self,
        allow_unused=True,
    ) -> LocalSpecification:
        """Generates the model's specification"""
        visitor = _ModelVisitor()
        visitor.visit(self, prefix=self.__prefix)
        statements = visitor.statements

        by_identifier = {
            s.definition.identifier: s
            for s in statements
            if s.definition.identifier
        }
        labels_by_identifier = {
            i: d.label for i, d in by_identifier.items() if d.label
        }
        formatter = _ModelFormatter(labels_by_identifier)
        reserved = {i.name: i for i in labels_by_identifier if i.name}
        rendered_by_title = collections.defaultdict(list)
        with global_formatting_scope(formatter, reserved):
            idens = set()
            for s in statements:
                rs = s.render()
                if not rs:
                    continue
                rendered_by_title[s.model.title].append(rs)
                if s.definition.identifier:
                    idens.add(s.definition.identifier)
            for iden in formatter.formatted_globals():
                if iden in idens:
                    continue
                ds = by_identifier.get(iden)
                if not ds:
                    raise Exception(f"Missing statement: {iden}")
                rs = ds.render()
                if not rs:
                    raise Exception(f"Missing rendered statement: {iden}")
                rendered_by_title[ds.model.title].append(rs)

        contents_by_title = {
            title: "".join(f"  {s} \\\\\n" for s in lines)
            for title, lines in rendered_by_title.items()
        }
        sources = [
            LocalSpecificationSource(
                title=title,
                text=f"$$\n\\begin{{align}}\n{contents}\\end{{align}}\n$$",
            )
            for title, contents in contents_by_title.items()
        ]
        spec = LocalSpecification(sources=sources)

        try:
            codes = ["ERR_UNUSED_DEFINITION"] if allow_unused else []
            spec = await spec.annotated(ignore_codes=codes)
        except Exception:
            _logger.warning("Unable to annotate specification", exc_info=True)
        return spec


class _ModelFormatter(IdentifierFormatter):
    def __init__(self, labels: Mapping[GlobalIdentifier, Label]) -> None:
        super().__init__(labels)

    def _format_dimension(self, label: Label, env: Environment) -> Name:
        i = _last_capital_index(label)
        if i is None:
            return label[0].upper()
        return f"{label[i]}^\\mathrm{{{label[:i]}}}" if i > 0 else label[i]

    def _format_parameter(self, label: Label, env: Environment) -> Name:
        i = _last_capital_index(label)
        if not i:
            return label[0].lower()
        return f"{label[i].lower()}^\\mathrm{{{label[:i]}}}"

    def _format_variable(self, label: Label, env: Environment) -> Name:
        i = _last_capital_index(label)
        r = label[i or 0].lower()
        g = _greek_letters.get(r, r)
        if not i:
            return g
        return f"{g}^\\mathrm{{{label[:i]}}}"

    def format_quantifier(
        self, identifier: QuantifierIdentifier, env: Environment
    ) -> Name:
        name = identifier.name
        if not name:
            sp = identifier.space
            if not isinstance(sp, Space):
                raise TypeError(f"Unexpected space: {sp}")
            if hasattr(sp, "identifier") and sp.identifier:  # Dimension
                name = _lower_principal(sp.identifier.format())
            else:  # Interval, possibly aliased
                group = identifier.outer_group
                for g in identifier.groups:
                    if g.rank == 1:
                        group = g
                        break
                if group:
                    name = _lower_principal(group.alias.format())
        return _first_available(name or _DEFAULT_QUANTIFIER_NAME, env)


_DEFAULT_QUANTIFIER_NAME = "x"


def _first_available(name: Name, env: Environment) -> Name:
    while name in env:
        name += "'"
    return name


def _last_capital_index(label: Label) -> Optional[int]:
    j = None
    for i, c in enumerate(label):
        if c.isupper():
            j = i
    return j


def _lower_principal(name: Name) -> Name:
    if "^" not in name:
        return name.lower()
    parts = name.split("^", 1)
    return f"{parts[0].lower()}^{parts[1]}"


_greek_letters = {
    "a": "\\alpha",
    "b": "\\beta",
    "c": "\\chi",
    "d": "\\delta",
    "e": "\\epsilon",
    "f": "\\phi",
    "g": "\\gamma",
    "h": "\\eta",
    "i": "\\iota",
    "j": "\\xi",  # TODO: Find better alternative
    "k": "\\kappa",
    "l": "\\lambda",
    "m": "\\mu",
    "n": "\\nu",
    "o": "\\omicron",
    "p": "\\pi",
    "q": "\\theta",
    "r": "\\rho",
    "s": "\\sigma",
    "t": "\\tau",
    "u": "\\psi",
    "v": "\\zeta",  # TODO: Find better alternative
    "w": "\\omega",
    "x": "\\xi",
    "y": "\\upsilon",
    "z": "\\zeta",
}
